package function.builtin.special;

import function.*;
import function.builtin.EVAL;
import sexpression.*;
import table.SymbolTable;

public class SETF extends LispFunction {

    private static SymbolTable symbolTable = new SymbolTable();

    public static SExpression lookupSymbolValue(String symbolName) {
        for (SymbolTable t = symbolTable; t != null; t = t.getParent())
            if (t.contains(symbolName))
                return t.get(symbolName);

        return null;
    }

    public static void setSymbolTable(SymbolTable newSymbolTable) {
        symbolTable = newSymbolTable;
    }

    public static SymbolTable getSymbolTable() {
        return symbolTable;
    }

    private ArgumentValidator argumentValidator;

    public SETF() {
        this.argumentValidator = new ArgumentValidator("SETF");
        this.argumentValidator.setExactNumberOfArguments(2);
        this.argumentValidator.setFirstArgumentExpectedType(Symbol.class);
    }

    public SExpression call(Cons argumentList) {
        argumentValidator.validate(argumentList);

        Cons cdr = (Cons) argumentList.getCdr();
        SExpression symbol = argumentList.getCar();
        SExpression value = EVAL.eval(cdr.getCar());

        SymbolTable table = findTableForSymbol(symbol);
        table.put(symbol.toString(), value);

        return value;
    }

    private SymbolTable findTableForSymbol(SExpression symbol) {
        SymbolTable table = symbolTable;

        while (!isSymbolInTable(symbol, table) && !isGlobalTable(table))
            table = table.getParent();

        return table;
    }

    private boolean isSymbolInTable(SExpression symbol, SymbolTable table) {
        return table.contains(symbol.toString());
    }

    private boolean isGlobalTable(SymbolTable table) {
        return table.getParent() == null;
    }

    public boolean evaluateArguments() {
        return false;
    }

}
